@@ -8,6 +8,7 @@
 # ----------------------------------------------------
 
 # %% Libraries
+import wget
 import os
 import numpy as np
 import pandas as pd
@@ -65,16 +66,18 @@ def df2onehot(df, dtypes='pandas', y_min=None, perc_min_num=None, hot_only=True,
     args['dtypes'] = dtypes
     args['verbose'] = verbose
     args['perc_min_num'] = perc_min_num
-    args['list_expand'] = list_expand
+    args['deep_extract'] = list_expand
     args['excl_background'] = excl_background
     labx = []
-    
+
     # Reset index
     df.reset_index(drop=True, inplace=True)
+    # Remove non-ascii chars
+    # df = _remove_non_ascii(df)
     # Determine Dtypes
-    df, dtypes = set_dtypes(df, args['dtypes'], is_list=args['list_expand'], perc_min_num=args['perc_min_num'], verbose=args['verbose'])
+    df, dtypes = set_dtypes(df, args['dtypes'], is_list=args['deep_extract'], perc_min_num=args['perc_min_num'], verbose=args['verbose'])
     # If any column is a list, also expand the list!
-    df, dtypes = _deep_extract(df, dtypes, args['verbose'])
+    df, dtypes = _deep_extract(df, dtypes, perc_min_num=args['perc_min_num'], verbose=args['verbose'])
 
     # Make empty frames
     out_numeric = pd.DataFrame()
@@ -195,8 +198,47 @@ def _concat(dftot, dfc):
                 dftot[colname] = dfc[colname].copy()
     return dftot
 
+def dict2df(dfc):
+    dftot = pd.DataFrame()
+    Iloc = ~dfc.isnull()
+    idx = np.where(Iloc)[0]
+    idxempty = list(np.where(Iloc==False)[0])
+
+    # result_list = [int(v) for k,v in dfc.iloc[i].items()]
+    # pd.DataFrame(data=dfc.iloc[i].items(), index=dfc.iloc[i].keys())
+
+    # Convert dict to dataframe
+    for i in idx:
+        # try:
+            # if isinstance(dict(), type(dfc.iloc[i])):
+            #     for key in dfc.iloc[i].keys():
+            #         print(key)
+            #         isinstance(dict, dfc.iloc[i][key])
+            #         pd.DataFrame.from_dict(dfc.iloc[i][key], orient='index')
+            # else:
+        dftmp = pd.DataFrame.from_dict(dfc.iloc[i], orient='index')
+        dftmp.rename(columns={0:i}, inplace=True)
+        # Combine into larger dataframe
+        dftot = pd.concat([dftot, dftmp], axis=1)
+        # except:
+        #     idxempty.append(i)
+
+    # Fill the empty ones with None
+    if not dftot.empty:
+        for i in idxempty:
+            dftmp = pd.DataFrame(index=dftot.index.values, data=[None]*len(dftot.index.values), columns=[i])
+            dftot = pd.concat([dftot, dftmp], axis=1)
+
+        # Transpose data and sort on index again
+        dftot = dftot.T
+        dftot.sort_index(inplace=True)
+        # Check
+        assert np.all(dftot.index.values==dfc.index.values)
+        
+    return(dftot, idxempty)
+
 # %%
-def _deep_extract(df, dtypes, verbose=3):
+def _deep_extract(df, dtypes, perc_min_num=None, verbose=3):
     # Check for any lists in dtypes
     Ilist = np.isin(dtypes,'list')
     Idict = np.isin(dtypes,'dict')
@@ -205,18 +247,26 @@ def _deep_extract(df, dtypes, verbose=3):
     idxrem1 = []
     idxrem2 = []
 
+
     # Expand dict
     if np.any(Idict):
+        if verbose >=3: print('[df2onehot] >\n[df2onehot] >Starting deep extraction..')
         idxCol = np.where(Idict)[0]
+        max_str_len = np.max(list(map(len, df.columns[idxCol].values.astype(str).tolist())))
         # Expand every columns that contains dict
         for idx in idxCol:
-            if verbose>=3: print('[df2onehot] >%s column is detected: [%s]' %(dtypes[idx], df.columns[idx]))
+            makespaces = ''.join([' '] * (max_str_len - len(df.columns[idx])))
             try:
-                dfc = pd.DataFrame.from_records(df.iloc[:,idx])
+                dfc, idxempty = dict2df(df.iloc[:, idx])
+                # dfc = pd.DataFrame.from_records(df.iloc[:,idx])
+                if verbose>=3: print('[df2onehot] >[%s]%s >deep extract > [%s]  [%d]' %(df.columns[idx], makespaces, dtypes[idx], dfc.shape[1]))
             except:
-                if verbose>=3: print('[df2onehot] >%s failed to process [%s].' %(dtypes[idx], df.columns[idx]))
-                dfc = df.iloc[:,idx].astype(str)
+                if verbose>=3: print('[df2onehot] >[%s]%s >deep extract > [failed]' %(df.columns[idx], makespaces))
+                # dfc = df.iloc[:,idx].astype(str)
+                # dfc = dfc.apply(_remove_non_ascii)
 
+            # Remove non ascii chars
+            # dfc = _remove_non_ascii(dfc)
             # Combine extracted columns into big dataframe
             dftot1 = pd.concat([dftot1, dfc], axis=1)
             # Add idx to remove
@@ -225,9 +275,10 @@ def _deep_extract(df, dtypes, verbose=3):
     # Expand list
     if np.any(Ilist):
         idxCol = np.where(Ilist)[0]
+        max_str_len = np.max(list(map(len, df.columns[idxCol].values.astype(str).tolist())))
         # Expand every columns that contains either list
         for idx in idxCol:
-            if verbose>=3: print('[df2onehot] >%s column is detected: [%s]' %(dtypes[idx], df.columns[idx]))
+            makespaces = ''.join([' '] * (max_str_len - len(df.columns[idx])))
             # Convert str/float/int to type
             df, uifeat = _col2type(df, dtypes, idx)
             # Convert column into onehot
@@ -236,63 +287,104 @@ def _deep_extract(df, dtypes, verbose=3):
             dftot2 = _concat(dftot2, dfc)
             # Add idx to remove
             idxrem2.append(idx)
+            if verbose>=3: print('[df2onehot] >[%s]%s >deep extract > [%s]  [%d]' %(df.columns[idx], makespaces, dtypes[idx], dfc.shape[1]))
 
     # Drop columns that are expanded
     idxrem = idxrem1+idxrem2
     if len(idxrem)>0:
-        df.drop(labels = df.columns[idxrem].values, axis=1, inplace=True)
+        # Remove the extracted column names from list and dict
         idxkeep = np.setdiff1d(np.arange(0, df.shape[1]), idxrem)
+        df.drop(labels = df.columns[idxrem].values, axis=1, inplace=True)
         dtypes = np.array(dtypes)
         dtypes = list(dtypes[idxkeep])
-        # Do the typing
+        # Combine the extracted list and dict data
         dftot = pd.concat([dftot1, dftot2], axis=1)
-        dftot, dtypes1 = set_dtypes(dftot, verbose=3)
+        # Remove repetative column names
+        dftot = _make_columns_unique(dftot, verbose=verbose)
+        # Set dtypes
+        dftot, dtypest = set_dtypes(dftot, perc_min_num=perc_min_num, verbose=3)
         # Combine into dataframe
         df = pd.concat([df, dftot], axis=1)
-        dtypes = dtypes + dtypes1
-        # Set dtypes for lists to bool
-        # dtypes = np.array(dtypes).astype('O')
-        # dtypes[idxrem2]='bool'
-        # dtypes = list(dtypes)
+        dtypes = dtypes + dtypest
+        if verbose>=3: print('[df2onehot] >[%d] additional columns extracted by deep extract.' %(dftot1.shape[1]+dftot2.shape[1]))
 
     # Return
-    if verbose>=3: print('[df2onehot] >[%d] additional columns extracted by deep extract.' %(dftot1.shape[1]+dftot2.shape[1]))
+    if df.shape[1]!=len(dtypes): raise Exception('[df2onehot] >Error: size of dtypes and dataframe does not match.')
     return(df, dtypes)
 
 
+# %% Remove repetative column
+def _make_columns_unique(dftot, verbose=3):
+    columns = dftot.columns.value_counts()
+    columns = columns[columns.values>1].index.values
+    if verbose>=3: print('[df2onehot] >[%d] repetative columns detected and a single one is taken: %s' %(len(columns), columns))
+    _, uiidx = np.unique(dftot.columns, return_index=True)
+    dftot = dftot.iloc[:,np.sort(uiidx)]
+    return dftot
+
 # %% Find columns
 def _findcol(x, cols):
+    # SLICE COPY WARNING!
     return(np.isin(cols, x))
 
-
-# %% Example data
-def import_example(getfile='titanic'):
-    """Import example.
-
+# %% Import example dataset from github.
+def import_example(data='titanic', url=None, sep=',', verbose=3):
+    """Import example dataset from github source.
+    
     Description
     -----------
+    Import one of the few datasets from github source or specify your own download url link.
 
     Parameters
     ----------
-    getfile : String, optional
-        'titanic'
+    data : str
+        Name of datasets: 'sprinkler', 'titanic', 'student', 'fifa', 'cancer', 'waterpump', 'retail'
+    url : str
+        url link to to dataset.
+    verbose : int, (default: 3)
+        Print message to screen.
 
     Returns
     -------
-    df : DataFrame
+    pd.DataFrame()
+        Dataset containing mixed features.
 
     """
-
-    if getfile=='titanic':
-        getfile='titanic_train.zip'
-
-    print('[df2onehot] >Loading %s..' %getfile)
-    curpath = os.path.dirname(os.path.abspath(__file__))
-    PATH_TO_DATA=os.path.join(curpath, 'data', getfile)
-    if os.path.isfile(PATH_TO_DATA):
-        df=pd.read_csv(PATH_TO_DATA, sep=',')
-        return df
+    if url is None:
+        if data=='sprinkler':
+            url='https://erdogant.github.io/datasets/sprinkler.zip'
+        elif data=='titanic':
+            url='https://erdogant.github.io/datasets/titanic_train.zip'
+        elif data=='student':
+            url='https://erdogant.github.io/datasets/student_train.zip'
+        elif data=='cancer':
+            url='https://erdogant.github.io/datasets/cancer_dataset.zip'
+        elif data=='fifa':
+            url='https://erdogant.github.io/datasets/FIFA_2018.zip'
+        elif data=='waterpump':
+            url='https://erdogant.github.io/datasets/waterpump/waterpump_test.zip'
+        elif data=='retail':
+            url='https://erdogant.github.io/datasets/marketing_data_online_retail_small.zip'
     else:
-        print('[df2onehot] >Oops! Example data not found!')
+        data = wget.filename_from_url(url)
+
+    if url is None:
+        if verbose>=3: print('[hnet] >Nothing to download.')
         return None
 
+    curpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
+    PATH_TO_DATA = os.path.join(curpath, wget.filename_from_url(url))
+    if not os.path.isdir(curpath):
+        os.makedirs(curpath, exist_ok=True)
+
+    # Check file exists.
+    if not os.path.isfile(PATH_TO_DATA):
+        if verbose>=3: print('[hnet] >Downloading [%s] dataset from github source..' %(data))
+        wget.download(url, curpath)
+
+    # Import local dataset
+    if verbose>=3: print('[hnet] >Import dataset [%s]' %(data))
+    df = pd.read_csv(PATH_TO_DATA, sep=sep)
+    # Return
+    return df
+
